@@ -10,11 +10,17 @@
 #define NODEJS_LTS_API 18
 #elif V8_MAJOR_VERSION == 9
 #define NODEJS_LTS_API 16
-#elif V8_MAJOR_VERSION == 8
+#elif V8_MAJOR_VERSION < 9
 #define NODEJS_LTS_API 14
 #endif
 #endif
 
+#if defined(NODEJS_LTS_API) && NODEJS_LTS_API < 18
+#define FixedArrayParam
+#else
+#define FixedArrayParam ,v8::Local<v8::FixedArray> import_arributes
+#endif
+
 #if V8_VERSION_TOTAL < 803
 #define PerformMicrotaskCheckpoint RunMicrotasks
 #endif
@@ -47,6 +53,17 @@ void ctx_finalizer(ctx_type* context ){
 static v8::Isolate* isolate = NULL;
 static v8::Platform* platformptr = NULL;
 
+static std::string read_text(std::string filename) {
+  if(filename.empty() || (filename.at(0) != '.' && filename.at(0) != '/'))
+    throw std::runtime_error("Invalid module: " + filename + " (paths should begin with . or /)");
+  std::ifstream t(filename);
+  if(t.fail())
+    throw std::runtime_error("Failed to open file: " + filename);
+  std::stringstream buffer;
+  buffer << t.rdbuf();
+  return buffer.str();
+}
+
 // Extracts a C string from a V8 Utf8Value.
 static const char* ToCString(const v8::String::Utf8Value& value) {
   return *value ? *value : "<string conversion failed>";
@@ -66,6 +83,98 @@ static void fatal_cb(const char* location, const char* message){
   REprintf("V8 FATAL ERROR in %s: %s", location, message);
 }
 
+static v8::Local<v8::Module> read_module(std::string filename, v8::Local<v8::Context> context);
+
+static v8::MaybeLocal<v8::Module> ResolveModuleCallback(v8::Local<v8::Context> context, v8::Local<v8::String> specifier
+                                                        FixedArrayParam, v8::Local<v8::Module> referrer) {
+  v8::String::Utf8Value name(context->GetIsolate(), specifier);
+  try {
+    return read_module(*name, context);
+  }
+  catch(const std::exception& err) {
+    isolate->ThrowException(ToJSString(err.what()));
+  }
+  return v8::Local<v8::Module>();
+};
+
+static v8::MaybeLocal<v8::Promise> dynamic_module_loader(v8::Local<v8::Context> context, v8::Local<v8::String> specifier) {
+  v8::Local<v8::Promise::Resolver> resolver = v8::Promise::Resolver::New(context).ToLocalChecked();
+  v8::MaybeLocal<v8::Promise> promise(resolver->GetPromise());
+  v8::String::Utf8Value name(context->GetIsolate(), specifier);
+  try {
+    v8::Local<v8::Module> module = read_module(*name, context);
+    resolver->Resolve(context, module->GetModuleNamespace()).FromMaybe(false);
+  } catch(const std::exception& err) {
+    resolver->Reject(context, ToJSString(err.what())).FromMaybe(false);
+  } catch(...) {
+    resolver->Reject(context, ToJSString("Unknown failure loading dynamic module")).FromMaybe(false);
+  }
+  return promise;
+}
+
+static v8::MaybeLocal<v8::Promise> ResolveDynamicModuleCallback(
+    v8::Local<v8::Context> context,
+#if V8_VERSION_TOTAL >= 908
+    v8::Local<v8::Data> host_defined_options,
+    v8::Local<v8::Value> resource_name,
+#elif V8_VERSION_TOTAL >= 603
+    v8::Local<v8::ScriptOrModule> referrer,
+#else
+    v8::Local<v8::String> referrer,
+#endif
+    v8::Local<v8::String> specifier
+    FixedArrayParam
+) {
+  return dynamic_module_loader(context, specifier);
+}
+
+static v8::ScriptOrigin make_origin(std::string filename){
+#if defined(NODEJS_LTS_API) && NODEJS_LTS_API < 18
+  return v8::ScriptOrigin(ToJSString( filename.c_str()), v8::Integer::New(isolate, 0),
+                          v8::Integer::New(isolate, 0), v8::False(isolate), v8::Local<v8::Integer>(),
+                          v8::Local<v8::Value>(), v8::False(isolate), v8::False(isolate), v8::True(isolate));
+#elif V8_VERSION_TOTAL < 1201
+  return v8::ScriptOrigin(isolate,ToJSString( filename.c_str()), 0, 0, false, -1,
+                          v8::Local<v8::Value>(), false, false, true);
+#else
+  return v8::ScriptOrigin(ToJSString( filename.c_str()), 0, 0, false, -1,
+                          v8::Local<v8::Value>(), false, false, true);
+#endif
+}
+
+static std::string throw_js_err(v8::Local<v8::Value> Exception, std::string filename){
+  std::string errmsg(std::string("Failed to import ES module '") + filename + "': " + *v8::String::Utf8Value(isolate, Exception));
+  throw std::runtime_error(errmsg);
+}
+
+/* Helper fun that compiles JavaScript source code */
+static v8::Local<v8::Module> read_module(std::string filename, v8::Local<v8::Context> context){
+  v8::Local<v8::String> source_text = ToJSString(read_text(filename).c_str());
+  if(source_text.IsEmpty())
+    throw std::runtime_error("Failed to read module file (check memory/stack limits.");
+  v8::TryCatch trycatch(isolate);
+  v8::ScriptCompiler::Source source(source_text, make_origin(filename));
+  v8::Local<v8::Module> module;
+  if (!v8::ScriptCompiler::CompileModule(isolate, &source).ToLocal(&module)){
+    if(trycatch.HasCaught())
+      throw_js_err(trycatch.Exception(), filename);
+    throw std::runtime_error("Failed to run CompileModule() source.");
+  }
+  if(!module->InstantiateModule(context, ResolveModuleCallback).FromMaybe(false)){
+    if(trycatch.HasCaught())
+      throw_js_err(trycatch.Exception(), filename);
+    throw std::runtime_error("Failed to run InstantiateModule().");
+  }
+  v8::Local<v8::Value> retValue;
+  if (!module->Evaluate(context).ToLocal(&retValue)){
+    if(trycatch.HasCaught())
+      throw_js_err(trycatch.Exception(), filename);
+    throw std::runtime_error("Failure loading module");
+  }
+  return module;
+}
+
+
 // [[Rcpp::init]]
 void start_v8_isolate(void *dll){
 #ifdef V8_ICU_DATA_PATH
@@ -103,6 +212,7 @@ void start_v8_isolate(void *dll){
   uintptr_t CurrentStackPosition = reinterpret_cast<uintptr_t>(__builtin_frame_address(0));
   isolate->SetStackLimit(CurrentStackPosition - kWorkerMaxStackSize);
 #endif
+  isolate->SetHostImportModuleDynamicallyCallback(ResolveDynamicModuleCallback);
 }
 
 /* Helper fun that compiles JavaScript source code */
